/*
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 */

#include    "mkind.h"

static	long	idx_gc;

static	int	check_mixsym ARGS((char *x,char *y));
static	int	compare ARGS((FIELD_PTR *a,FIELD_PTR *b));
static	int	compare_one ARGS((char *x,char *y));
static	int	compare_page ARGS((FIELD_PTR a,FIELD_PTR b));
static	int	compare_string ARGS((unsigned char *a,unsigned char *b));
static	int	new_strcmp ARGS((unsigned char *a, unsigned char *b,
				 int option));
static	void	print_keys ARGS((const char *title));

void
sort_idx(VOID_ARG)
{
    MESSAGE("Sorting entries...%s", "");

    if (debug)
	print_keys("\nDEBUG: sort_idx(): keys before sort\n");

    idx_dc = 0;
    idx_gc = 0L;
    qqsort((char *) idx_key, (int) idx_gt, (int) sizeof(FIELD_PTR),
	(int (*) ARGS((char*,char*)))compare);
    MESSAGE("done (%ld comparisons).\n", idx_gc);

    if (debug)
	print_keys("DEBUG: sort_idx(): keys after sort\n");
}

static int
#if STDC
compare(FIELD_PTR *a, FIELD_PTR *b)
#else
compare(a, b)
FIELD_PTR *a;
FIELD_PTR *b;
#endif
{
    int     	i;
    int     	dif = 0;		/* loop might not set this */
    FIELD_PTR	ap = *a;		/* dereference for access speed */
    FIELD_PTR	bp = *b;

    idx_gc++;
    IDX_DOT(CMP_MAX);

    for (i = 0; i < FIELD_MAX; i++)
    {
	/* NB: This loop is optimized to do inline tests for the
	   common case of empty fields.  Without these checks,
	   compare_one() is the third most frequently called function
	   in MakeIndex, after tolower() and group_type().  In tests
	   with the three large book indexes {book,heck,texbk}.idx in
	   ../test/, compare() is called 17.2 times for each output
	   index entry.  Before the optimizations here, compare_one()
	   was called 31.9 times for each output index entry; the
	   optimizations reduced that value to 20.8.  */

	if ((ap->sf[i][0] == NUL) && (bp->sf[i][0] == NUL))
	    continue;			/* fast test for empty fields */

	/* compare the sort fields */
	if ((dif = compare_one(ap->sf[i], bp->sf[i])) != 0)
	    break;

	if ((ap->af[i][0] == NUL) && (bp->af[i][0] == NUL))
	    continue;			/* fast test for empty fields */

	/* compare the actual fields */
	if ((dif = compare_one(ap->af[i], bp->af[i])) != 0)
	    break;
    }

    /* both key aggregates are identical, compare page numbers */
    if (i == FIELD_MAX)
	dif = compare_page(ap, bp);
    return (dif);
}

int
#if STDC
compare_number(FIELD_PTR a,FIELD_PTR b,page_count i)
#else
compare_number(a,b,i)
FIELD_PTR a;
FIELD_PTR b;
page_count i;
#endif
{
    /* Compare i-th page numbers in a and b, return -1 (a < b), 0 (a == b),
       or +1 (a > b).  Normally, this is just a straightforward comparison
       of numerical values of npg[] entries.  However, in the case of mixed
       Arabic and Roman numbers, like 120A, 120B, 120C, ..., we first do
       a numerical comparison, and then if the numbers match, order by
       string comparison. */

    if (PAGE_VALUE(a,i) < PAGE_VALUE(b,i))
	return (-1);
    else if (PAGE_VALUE(a,i) == PAGE_VALUE(b,i))
    {
	if (PAGE_TYPE(a,i) == ARABALPHA)
	{
	    if (PAGE_TYPE(b,i) == ARABALPHA) /* 120? vs 120? */
		return (compare_string((unsigned char*)PAGE_STRING(a,i),
				       (unsigned char*)PAGE_STRING(b,i)));
	    else			/* 120C vs 120 */
		return (1);
	}
	else if (PAGE_TYPE(b,i) == ARABALPHA) /* 120 vs 120C */
	    return (-1);
	else
	    return (0);
    }
    else
	return (1);
}

static int
#if STDC
compare_one(char *x,char *y)
#else
compare_one(x, y)
char   *x;
char   *y;
#endif
{
    int     m;
    int     n;

#if 0	/* this test was moved outside calls to compare_one() in */
	/* MakeIndex 2.13; do NOT remove this code! */
    if ((x[0] == NUL) && (y[0] == NUL))
	return (0);
#endif

    if (x[0] == NUL)
	return (-1);

    if (y[0] == NUL)
	return (1);

    m = group_type(x);
    n = group_type(y);

    /* both pure digits */
    if ((m >= 0) && (n >= 0))
	return (m - n);

    /* x digit, y non-digit */
    if (m >= 0) {
	if (german_sort)
	    return (1);
	else
	    return ((n == -1) ? 1 : -1);
    }
    /* x non-digit, y digit */
    if (n >= 0) {
	if (german_sort)
	    return (-1);
	else
	    return ((m == -1) ? -1 : 1);
    }
    /* strings started with a symbol (including digit) */
    if ((m == SYMBOL) && (n == SYMBOL))
	return (check_mixsym(x, y));

    /* x symbol, y non-symbol */
    if (m == SYMBOL)
	return (-1);

    /* x non-symbol, y symbol */
    if (n == SYMBOL)
	return (1);

    /* strings with a leading letter, the ALPHA type */
    return (compare_string((unsigned char*)x, (unsigned char*)y));
}

static int
#if STDC
check_mixsym(char *x, char *y)
#else
check_mixsym(x, y)
char   *x;
char   *y;
#endif
{
    int     m;
    int     n;

    m = ISDIGIT(x[0]);
    n = ISDIGIT(y[0]);

    if (m && !n)
	return (1);

    if (!m && n)
	return (-1);

    return (strcmp(x, y));
}


static int
#if STDC
compare_string(unsigned char *a, unsigned char *b)
#else
compare_string(a, b)
unsigned char   *a;
unsigned char   *b;
#endif
{
    int     i = 0;
    int     j = 0;
    int     al;
    int     bl;

    while ((a[i] != NUL) || (b[j] != NUL)) {
	if (a[i] == NUL)
	    return (-1);
	if (b[j] == NUL)
	    return (1);
	if (letter_ordering) {
	    if (a[i] == SPC)
		i++;
	    if (b[j] == SPC)
		j++;
	}
	al = TOLOWER(a[i]);
	bl = TOLOWER(b[j]);

	if (al != bl)
	    return (al - bl);
	i++;
	j++;
    }
    if (german_sort)
	return (new_strcmp(a, b, GERMAN));
    else
	return (STRCMP(a,b));
}

static int
#if STDC
compare_page(FIELD_PTR a, FIELD_PTR b)
#else
compare_page(a, b)
FIELD_PTR a;
FIELD_PTR b;
#endif
{
    int     	m = 0;
    page_count	i = 0;

    while ((i < a->count) && (i < b->count) &&
	   (compare_number(a,b,i) == 0))
	i++;
    if ((i < a->count) && (i < b->count))
	m = compare_number(a,b,i);

    if (m == 0)
    {				/* common leading page numbers match */
	if ((i == a->count) && (i == b->count))
	{			/* all page numbers match */
	    /***********************************************************
	    NB: entire loop body rewritten for MakeIndex 2.11.
	    We have identical entries, except possibly in encap fields.
	    The ordering is tricky here.  Consider the following input
	    sequence of index names, encaps, and page numbers:

		foo|(	2
		foo|)	6
		foo|(	6
		foo|)	10

	    This might legimately occur when a page range ends, and
	    subsequently, a new range starts, on the same page.  If we
	    just order by range_open and range_close (here, parens),
	    then we will produce

		foo|(	2
		foo|(	6
		foo|)	6
		foo|)	10

	    This will later generate the index entry

		foo, 2--6, \({6}, 10

	    which is not only wrong, but has also introduced an illegal
	    LaTeX macro, \({6}, because the merging step treated this
	    like a \see{6} entry.

	    The solution is to preserve the original input order, which
	    we can do by treating range_open and range_close as equal,
	    and then ordering by input line number.  This will then
	    generate the correct index entry

		foo, 2--10

	    Ordering inconsistencies from missing range open or close
	    entries, or mixing roman and arabic page numbers, will be
	    detected later.
	    ***********************************************************/

#define isrange(c) ( ((c) == idx_ropen) || ((c) == idx_rclose) )

	    /* Order two range values by input line number */

	    if (isrange(*a->encap) && isrange(*b->encap))
		m = (int)a->lc - (int)b->lc;

	    /* Handle identical encap fields; neither is a range delimiter */

	    else if (STREQ(a->encap, b->encap))
	    {
		/* If neither are yet marked duplicate, mark the second
		of them to be ignored. */
		if ((PAGE_TYPE(a,0) != DUPLICATE) &&
		    (PAGE_TYPE(b,0) != DUPLICATE))
		    PAGE_TYPE(b,0) = DUPLICATE;
		/* leave m == 0 to show equality */
	    }

	    /* Encap fields differ: only one may be a range delimiter, */
	    /* or else neither of them is.   If either of them is a range */
	    /* delimiter, order by input line number; otherwise, order */
	    /* by name. */

	    else
	    {
		if ( isrange(*a->encap) || isrange(*b->encap) )
					/* order by input line number */
		    m = (int)a->lc - (int)b->lc;
		else			/* order non-range items by */
					/* their encap strings */
		    m = compare_string((unsigned char*)(a->encap),
				       (unsigned char*)(b->encap));
	    }
	}
	else if ((i == a->count) && (i < b->count))
	    m = -1;
	else if ((i < a->count) && (i == b->count))
	    m = 1;
    }
    return (m);
}


static int
#if STDC
new_strcmp(unsigned char *s1, unsigned char *s2, int option)
#else
new_strcmp(s1, s2, option)
unsigned char   *s1;
unsigned char   *s2;
int     option;
#endif
{
    int     i;

    i = 0;
    while (s1[i] == s2[i])
	if (s1[i++] == NUL)
	    return (0);
    if (option)			       /* ASCII */
	return (ISUPPER(s1[i]) ? -1 : 1);
    else			       /* GERMAN */
	return (ISUPPER(s1[i]) ? 1 : -1);
}

static void
#if STDC
print_keys(const char *title)
#else
print_keys(title)
const char	*title;
#endif
{
    entry_number	k;
    page_count		m;

    (void)fprintf(stddbg,"%s",title);
    for (k = 0; k < idx_gt; ++k)
    {
	(void)fprintf(stddbg, "\tidx_key[%ld]\t[%-10s][%4s]\n",
		      (long)k, idx_key[k]->sf[0], idx_key[k]->lpg);
	for (m = 0; m < idx_key[k]->count; ++m)
	    (void)fprintf(stddbg, "\t\t\t(pno=%ld type=%d page=[%s])\n",
			  (long)(PAGE_VALUE(idx_key[k],m)),
			  (int)(PAGE_TYPE(idx_key[k],m)),
			  PAGE_STRING(idx_key[k],m));
    }
}
